/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _SharedObjectFromKernel_instances, _SharedObjectFromKernel_lazyData, _SharedObjectFromKernel_kernelArgs, _SharedObjectFromKernel_initializeData, _SharedObjectFromKernel_kernel_get;
import { assert, fail } from "@fluidframework/core-utils/internal";
import { createSharedObjectKind, SharedObject, } from "./sharedObject.js";
/**
 * SharedObject implementation that delegates to a SharedKernel.
 * @typeParam TOut - The type of the object exposed to the app.
 * Once initialized, instances of this class forward properties to the `TOut` value provided by the factory.
 * See {@link mergeAPIs} for more limitations.
 *
 * @remarks
 * The App facing API (TOut) needs to be implemented by this object which also has to implement the runtime facing API (ISharedObject).
 *
 * Requiring both of these to be implemented by the same object adds some otherwise unnecessary coupling.
 * This class is a workaround for that, which takes separate implementations of the two APIs and merges them into one using {@link mergeAPIs}.
 */
class SharedObjectFromKernel extends SharedObject {
    constructor(id, runtime, attributes, factory, telemetryContextPrefix) {
        super(id, runtime, attributes, telemetryContextPrefix);
        _SharedObjectFromKernel_instances.add(this);
        this.factory = factory;
        /**
         * Lazy init here so kernel can be constructed in loadCore when loading from existing data.
         *
         * Explicit initialization to undefined is done so Proxy knows this property is from this class (via `Reflect.has`),
         * not from the grafted APIs.
         */
        _SharedObjectFromKernel_lazyData.set(this, undefined);
        _SharedObjectFromKernel_kernelArgs.set(this, void 0);
        __classPrivateFieldSet(this, _SharedObjectFromKernel_kernelArgs, {
            sharedObject: this,
            serializer: this.serializer,
            submitLocalMessage: (op, localOpMetadata) => this.submitLocalMessage(op, localOpMetadata),
            eventEmitter: this,
            logger: this.logger,
            idCompressor: runtime.idCompressor,
            lastSequenceNumber: () => this.deltaManager.lastSequenceNumber,
        }, "f");
    }
    summarizeCore(serializer, telemetryContext, incrementalSummaryContext) {
        return __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "a", _SharedObjectFromKernel_kernel_get).summarizeCore(serializer, telemetryContext, incrementalSummaryContext);
    }
    initializeLocalCore() {
        __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "m", _SharedObjectFromKernel_initializeData).call(this, this.factory.create(__classPrivateFieldGet(this, _SharedObjectFromKernel_kernelArgs, "f")));
    }
    async loadCore(storage) {
        __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "m", _SharedObjectFromKernel_initializeData).call(this, await this.factory.loadCore(__classPrivateFieldGet(this, _SharedObjectFromKernel_kernelArgs, "f"), storage));
    }
    onDisconnect() {
        __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "a", _SharedObjectFromKernel_kernel_get).onDisconnect();
    }
    reSubmitCore(content, localOpMetadata) {
        __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "a", _SharedObjectFromKernel_kernel_get).reSubmitCore(content, localOpMetadata);
    }
    applyStashedOp(content) {
        __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "a", _SharedObjectFromKernel_kernel_get).applyStashedOp(content);
    }
    processCore() {
        fail("processCore should not be called");
    }
    processMessagesCore(messagesCollection) {
        __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "a", _SharedObjectFromKernel_kernel_get).processMessagesCore(messagesCollection);
    }
    rollback(content, localOpMetadata) {
        if (__classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "a", _SharedObjectFromKernel_kernel_get).rollback === undefined) {
            super.rollback(content, localOpMetadata);
        }
        else {
            __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "a", _SharedObjectFromKernel_kernel_get).rollback(content, localOpMetadata);
        }
    }
    didAttach() {
        __classPrivateFieldGet(this, _SharedObjectFromKernel_instances, "a", _SharedObjectFromKernel_kernel_get).didAttach?.();
    }
}
_SharedObjectFromKernel_lazyData = new WeakMap(), _SharedObjectFromKernel_kernelArgs = new WeakMap(), _SharedObjectFromKernel_instances = new WeakSet(), _SharedObjectFromKernel_initializeData = function _SharedObjectFromKernel_initializeData(data) {
    assert(__classPrivateFieldGet(this, _SharedObjectFromKernel_lazyData, "f") === undefined, "initializeData must be called first and only once");
    __classPrivateFieldSet(this, _SharedObjectFromKernel_lazyData, data, "f");
    // Make `this` implement TOut.
    mergeAPIs(this, data.view);
}, _SharedObjectFromKernel_kernel_get = function _SharedObjectFromKernel_kernel_get() {
    return (__classPrivateFieldGet(this, _SharedObjectFromKernel_lazyData, "f") ?? fail("must initializeData first")).kernel;
};
/**
 * When present on a method, it indicates the methods return value should be replaced with `this` (the wrapper)
 * when wrapping the object with the method.
 * @remarks
 * This is useful when using {@link mergeAPIs} with methods where the return type is `this`, like `Map.set`.
 * @internal
 */
export const thisWrap = Symbol("selfWrap");
/**
 * Add getters to `base` which forward own properties from `extra`.
 * @remarks
 * This only handles use of "get" and "has".
 * Therefore, APIs involving setting properties should not be used as `Extra`.
 *
 * Functions from `extra` are bound to the `extra` object and support {@link thisWrap}.
 *
 * Asserts when properties collide.
 * @internal
 */
export function mergeAPIs(base, extra) {
    for (const [key, descriptor] of Object.entries(Object.getOwnPropertyDescriptors(extra))) {
        assert(!Reflect.has(base, key), "colliding properties");
        // Detect and special case functions.
        // Currently this is done eagerly (when mergeAPIs is called) rather than lazily (when the property is read):
        // this eager approach should result in slightly better performance,
        // but if functions on `extra` are reassigned over time it will produce incorrect behavior.
        // If this functionality is required, the design can be changed.
        let getter;
        // Bind functions to the extra object and handle thisWrap.
        if (typeof descriptor.value === "function") {
            // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
            const fromExtra = descriptor.value;
            getter = () => forwardMethod(fromExtra, extra, base);
            // To catch (and error on) cases where the function is reassigned and this eager binding approach is not appropriate, make it non-writable.
            Object.defineProperty(extra, key, { ...descriptor, writable: false });
        }
        else {
            getter = () => extra[key];
        }
        Object.defineProperty(base, key, {
            configurable: false,
            enumerable: descriptor.enumerable,
            get: getter,
            // If setters become required, support them here.
        });
    }
}
/**
 * Wrap a method `f` of `oldThis` to be a method of `newThis`.
 * @remarks
 * The wrapped function will be called with `oldThis` as the `this` parameter.
 * It also accounts for when `f` is marked with {@link thisWrap}.
 */
function forwardMethod(f, oldThis, newThis) {
    // eslint-disable-next-line unicorn/prefer-ternary
    if (thisWrap in f) {
        return (...args) => {
            const result = f.call(oldThis, ...args);
            assert(result === oldThis, "methods returning thisWrap should return this");
            return newThis;
        };
    }
    else {
        return f.bind(oldThis);
    }
}
/**
 * Utility to create a {@link @fluidframework/datastore-definitions#IChannelFactory} classes.
 * @remarks
 * Use {@link makeSharedObjectKind} instead unless exposing the factory is required for legacy API compatibility.
 * @internal
 */
// eslint-disable-next-line @typescript-eslint/explicit-function-return-type
function makeChannelFactory(options) {
    class ChannelFactory {
        /**
         * {@inheritDoc @fluidframework/datastore-definitions#IChannelFactory."type"}
         */
        get type() {
            return ChannelFactory.Type;
        }
        /**
         * {@inheritDoc @fluidframework/datastore-definitions#IChannelFactory.attributes}
         */
        get attributes() {
            return ChannelFactory.Attributes;
        }
        /**
         * {@inheritDoc @fluidframework/datastore-definitions#IChannelFactory.load}
         */
        async load(runtime, id, services, attributes) {
            const shared = new SharedObjectFromKernel(id, runtime, attributes, options.factory, options.telemetryContextPrefix);
            await shared.load(services);
            return shared;
        }
        /**
         * {@inheritDoc @fluidframework/datastore-definitions#IChannelFactory.create}
         */
        create(runtime, id) {
            const shared = new SharedObjectFromKernel(id, runtime, ChannelFactory.Attributes, options.factory, options.telemetryContextPrefix);
            shared.initializeLocal();
            return shared;
        }
    }
    /**
     * {@inheritDoc @fluidframework/datastore-definitions#IChannelFactory."type"}
     */
    ChannelFactory.Type = options.type;
    /**
     * {@inheritDoc @fluidframework/datastore-definitions#IChannelFactory.attributes}
     */
    ChannelFactory.Attributes = options.attributes;
    return ChannelFactory;
}
/**
 * Utility to create a {@link SharedObjectKind}.
 * @privateRemarks
 * Using this API avoids having to subclasses any Fluid Framework types,
 * reducing the coupling between the framework and the SharedObject implementation.
 * @internal
 */
export function makeSharedObjectKind(options) {
    return createSharedObjectKind(makeChannelFactory(options));
}
//# sourceMappingURL=sharedObjectKernel.js.map